from typing import Iterable, Callable, Any

from davidkhala.utils.http_request import Request


class API(Request):
    def __init__(self, api_key: str, base_url="https://api.dify.ai/v1"):
        super().__init__({'bearer': api_key})
        self.base_url = base_url
        self.api_key = api_key


class Iterator(Iterable):
    def __iter__(self):
        return self

    def __init__(self, get_fn: Callable[[int, int, str], Any], *, size=20, keyword=None):
        self.last_response = None
        self.keyword = keyword
        self.size = size
        self.fn = get_fn

    def __next__(self):
        if self.last_response and not self.last_response['has_more']:
            raise StopIteration
        page = 1 if not self.last_response else self.last_response['page'] + 1
        self.last_response = self.fn(page, self.size, self.keyword)
        return self.last_response['data']



from datetime import datetime
from time import sleep
from typing import Callable

from davidkhala.ai.agent.dify.api.knowledge.model import DocMetadataModel, MetadataType
from davidkhala.ai.agent.dify.console.session import ConsoleDerived, ConsoleUser
from davidkhala.ai.agent.dify.const import IndexingStatus
from davidkhala.ai.agent.dify.interface import IndexingError


class Operation(ConsoleDerived):
    def __init__(self, context: ConsoleUser, dataset: str):
        super().__init__(context)
        self.base_url = f"{context.base_url}/datasets/{dataset}"
        self.dataset = dataset


class DocumentOperation(Operation):

    def website_sync(self, document: str, *, wait_until=True):
        """
        for dataset with runtime_mode=standard
        cannot be used towards a pipeline dataset. Otherwise, you will see error "no website import info found"
        It can be used to force document into error status
        """

        r = self.request(f"{self.base_url}/documents/{document}/website-sync", "GET")
        assert r == {"result": "success"}
        if wait_until:
            return self.wait_until(document)
        return None

    def retry(self, *documents: str, wait_until=True):
        """
        It cannot trigger rerun on success documents
        """
        url = f"{self.base_url}/retry"
        self.request(url, "POST", json={
            'document_ids': documents,
        })
        # response status code will be 204
        if wait_until:
            return [self.wait_until(document) for document in documents]
        return None

    def rerun(self, *documents: str):
        for document in documents:
            try:
                self.website_sync(document)
                assert False, "expect IndexingError"
            except IndexingError:
                pass
        return self.retry(*documents)

    def wait_until(self, document: str, *,
                   expect_status=None,
                   from_status=None,
                   interval=1
                   ):
        # TODO this has public api version in https://docs.dify.ai/api-reference/documents/get-document-embedding-status-progress
        if not expect_status:
            expect_status = [IndexingStatus.FAILED, IndexingStatus.COMPLETED]
        url = f"{self.base_url}/documents/{document}/indexing-status"
        if from_status is None:
            from_status = [IndexingStatus.WAITING, IndexingStatus.PARSING]
        r = self.request(url, "GET")
        status = r['indexing_status']
        assert status in from_status, f"current status: {status}, expect: {from_status}"
        while status not in expect_status:
            sleep(interval)
            r = self.request(url, "GET")
            status = r['indexing_status']
        if status == IndexingStatus.FAILED: raise IndexingError(r['error'])
        return r


class DocumentMetadata(Operation):
    def list(self) -> list[DocMetadataModel]:
        r = self.request(f"{self.base_url}/metadata", "GET")
        return [DocMetadataModel.model_validate(_) for _ in r['doc_metadata']]

    def add(self, _type: MetadataType, name: str) -> str:
        r = self.request(f"{self.base_url}/metadata", "POST", json={
            'type': _type, 'name': name
        })
        return r['id']

    def delete(self, metadata_id: str):
        self.request(f"{self.base_url}/metadata/{metadata_id}", "DELETE")

    def set(self, *documents: str, metadata: dict):
        url = f"{self.base_url}/documents/metadata"
        metadata_list = []
        datasets = self.list()

        for name, value in metadata.items():
            _type: MetadataType
            match value:
                case str():
                    _type = "string"
                case datetime():
                    _type = "time"
                    value = int(value.timestamp())
                case _:
                    _type = "number"
            _id = next((_.id for _ in datasets if _.name == name), None)
            if not _id:
                _id = self.add(_type, name)

            metadata_list.append({
                'id': _id,
                'type': _type,
                'name': name, 'value': value
            })
        self.request(url, "POST", json={
            'operation_data': [
                {'document_id': document, 'metadata_list': metadata_list, "partial_update": False}
                for document in documents
            ],
        })



from davidkhala.ai.agent.dify.const import IndexingStatus
from davidkhala.ai.agent.dify.model import ID


class Document(ID):
    position: int
    data_source_type: str
    data_source_info: dict[str, str|int]
    name: str
    indexing_status: IndexingStatus
    error: str | None
    enabled: bool


class Dataset(ID):
    name: str
    description: str



from __future__ import annotations

from typing import Iterable, Literal

import requests

from davidkhala.ai.agent.dify.api import API, Iterator
from davidkhala.ai.agent.dify.api.knowledge.dataset import Dataset
from davidkhala.ai.agent.dify.api.knowledge.model import ChunkDict, DocumentModel, MetadataDocumentModel, \
    NonMetadataDocumentModel


class Document(API):
    def __init__(self, d: Dataset.Instance, document_id: str):
        super().__init__(d.api_key, f"{d.base_url}/documents/{document_id}")

    def exist(self):
        try:
            self.get()
            return True
        except requests.exceptions.HTTPError as e:
            if e.response.status_code == 404:
                return False
            else:
                raise e

    def get(self,
            metadata: Literal['all', 'only', 'without'] = 'all'
            ) -> DocumentModel | MetadataDocumentModel | NonMetadataDocumentModel:
        """
        :param metadata:
            - 'only' returns metadata
            - 'without' returns anything without metadata
        """

        r = self.request(self.base_url, "GET", params={
            'metadata': metadata
        })
        match metadata:
            case 'only':
                return MetadataDocumentModel.model_validate(r)
            case 'without':
                return NonMetadataDocumentModel.model_validate(r)

        return DocumentModel.model_validate(r)

    def paginate_chunks(self, page=1, size=20, keyword=None):
        assert 0 < size < 101
        return self.request(f"{self.base_url}/segments", "GET", params={
            'page': page,
            'limit': size,
            'keyword': keyword
        })

    def list_chunks(self, keyword=None) -> Iterable[ChunkDict]:
        for chunk_batch in Iterator(self.paginate_chunks, keyword=keyword):
            for chunk in chunk_batch:
                yield chunk

    def delete(self):
        if self.exist():
            self.request(self.base_url, "DELETE")
